/*
mdl.y - bison source for mdl (Population Model)
Release 1.0.0
Copyright (C) 2001-2004  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/*
Usage:
    	see makefile for compile instructions
*/

%{

/* Set YYDEBUG to 1 to insert debugging code */
/* and to turn on debugging in bison */
/* Run flex with the -d option to turn on flex debugging */

#define YYDEBUG 0

#define MAXYEAR 1024
#define VERBOSE 2
#define YYSTYPE char *
#define DCML '.'

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <values.h>
#include "mdl.tab.h"

int yydebug = YYDEBUG;
int dbsw; /* debug switch */
int errsw;
int maxage;
int maxyrs;
int frstage;
int lastage;
double startpop;
double children;
double minattr;
double maxattr;
double cbattr;
double oldageattr;
double prevftot;
double prevmtot;
double cbspan;
double cbrate;

%}

%token MAXAGE MAXYRS STARTPOP FRSTCBAGE
%token LASTCBAGE CHILDREN MINATTR MAXATTR CBATTR
%token OLDAGEATTR
%token POSFLOAT NUM

%%

fldlst:		fldexp
		= {
		if (dbsw)
		   fprintf(stderr,"first field done\n");
		}
		| fldlst fldexp
		= {
		if (dbsw)
		   fprintf(stderr,"next field done\n");
		}
		;

fldexp:		MAXAGE '=' NUM
		= {
		maxage = atoi(yylval);
		if (maxage < 1 || maxage > 120)
		   {
		   fprintf(stderr,"\nInvalid "
		      "maximum age <%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\nmax_age = %d\n",
		      maxage);
		}

		| MAXYRS '=' NUM
		= {
		maxyrs = atoi(yylval);
		if (maxyrs < 1 || maxyrs > 999999999)
		   {
		   fprintf(stderr,"\nInvalid "
		      "maximum years <%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\nmax_years = %d\n",
		      maxyrs);
		}

		| STARTPOP '=' NUM
		= {
		unsigned int i;
		i = atoi(yylval);
		startpop = (double) i;
		if (startpop < 1 || startpop > 1000000000)
		   {
		   fprintf(stderr,"\nInvalid "
		      "starting population <%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\nstarting_pop = %f\n",
		      startpop);
		}

		| FRSTCBAGE '=' NUM
		= {
		frstage = atoi(yylval);
		if (frstage < 8 || frstage > 40)
		   {
		   fprintf(stderr,"\nInvalid "
		      "first childbirth "
		      "age <%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\nfirst_cb_age = %d\n",
		      frstage);
		}

		| LASTCBAGE '=' NUM
		= {
		lastage = atoi(yylval);
		if (lastage < 8 || lastage > 40)
		   {
		   fprintf(stderr,"\nInvalid "
		      "last childbirth "
		      "age <%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\nlast_cb_age = %d\n",
		      lastage);
		}

		| CHILDREN '=' POSFLOAT
		= {
		children = atof(yylval);
		if (children < 0.0000000001
		   || children > 25.0)
		   {
		   fprintf(stderr,"\nInvalid "
		      "number of children <%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\nchildren = %f\n",
		      children);
		}

		| MINATTR '=' POSFLOAT
		= {
		double tmpattr;
		tmpattr = atof(yylval);
		minattr = 1.0 - tmpattr;
		if (minattr < 0.0000000001
		   || minattr > 0.999999999)
		   {
		   fprintf(stderr,"\nInvalid "
		      "minimum attrition "
		      "<%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\nmin_attrition = %f\n",
		      tmpattr);
		}

		| MAXATTR '=' POSFLOAT
		= {
		double tmpattr;
		tmpattr = atof(yylval);
		maxattr = 1.0 - tmpattr;
		if (maxattr < 0.0000000001
		   || maxattr > 0.999999999)
		   {
		   fprintf(stderr,"\nInvalid "
		      "maximum attrition "
		      "<%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\nmax_attrition = %f\n",
		      tmpattr);
		}

		| CBATTR '=' POSFLOAT
		= {
		double tmpattr;
		tmpattr = atof(yylval);
		cbattr = 1.0 - tmpattr;
		if (cbattr < 0.0000000001
		   || cbattr > 0.999999999)
		   {
		   fprintf(stderr,"\nInvalid "
		      "childbirth attrition "
		      "<%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\ncb_attrition = %f\n",
		      tmpattr);
		}

		| OLDAGEATTR '=' POSFLOAT
		= {
		double tmpattr;
		tmpattr = atof(yylval);
		oldageattr = 1.0 - tmpattr;
		if (oldageattr < 0.0000000001
		   || oldageattr > 0.999999999)
		   {
		   fprintf(stderr,"\nInvalid "
		      "old age attrition "
		      "<%s>\n",
		      yylval);
		   errsw = 1;
		   }
		if (dbsw)
		   fprintf(stderr,"\nold_age_attrition = %f\n",
		      tmpattr);
		}
		;

%%

void putstx(pgm)
char *pgm;
   {
   fprintf(stderr,"Usage: %s\n", pgm);
   exit(1);
   }

yyerror(msg)
   {
   fprintf(stderr,"%s\n", msg);
   exit(1);
   } /* yyerror */

void shwcfg()
   {
   printf("dbsw     = %d\n", dbsw);
   printf("errsw    = %d\n", errsw);
   printf("max_cb_age       = %d\n", maxage);
   printf("max_years        = %d\n", maxyrs);
   printf("first_cb_age     = %d\n", frstage);
   printf("last_cb_age      = %d\n", lastage);
   printf("children         = %f\n", children);
   printf("starting_pop     = %f\n", startpop);
   printf("min_attrition    = %f\n", minattr);
   printf("max_attrition    = %f\n", maxattr);
   } /* shwcfg */

void shwpop(gen,m,f)
int gen;
double m[],f[];
   {
   int i;
   double totm,totf;
   double mdiff,fdiff;
   i = 0;
   totm = totf = 0.0;
   printf("         Year %d\n\n", gen);
   printf("          Age           Males "
      "        Females           Total\n");
   printf("          ---           ----- "
      "        -------           -----\n");
   while (m[i] > 2.0 && f[i] > 2.0)
      {
      printf("       %6d %15.0f %15.0f %15.0f\n",
         i, m[i], f[i], m[i]+f[i]);
      totm += m[i];
      totf += f[i];
      i++;
      }
   putchar('\n');
   printf("       %s %15.0f %15.0f %15.0f\n",
      " Total", totm, totf, totm+totf);
   mdiff = totm - prevmtot;
   fdiff = totf - prevftot;
   putchar('\n');
   printf("       %s %15.0f %15.0f %15.0f\n",
      "Change", mdiff, fdiff, mdiff+fdiff);
   printf("\f\n");
   prevmtot = totm;
   prevftot = totf;
   } /* shwpop */

void genmdl()
   {
   int gen,i;
   double pop,attr;
   double sinrange,dx,x;
   double m[MAXYEAR],f[MAXYEAR];
   for (i=0;i<MAXYEAR;i++)
      m[i] = f[i] = 0.0;
   pop = startpop / maxage;
   cbspan = (double) lastage - frstage + 1.0;
   cbrate = children / cbspan;
   sinrange = minattr - maxattr;
   attr = maxattr;
   dx = M_PI / maxage;
   x = 0;
   i = 0;
   while (pop > 2.0)
      {
      if (x < M_PI)
         attr = (sinrange * sin(x)) + maxattr;
      pop *= attr;
      m[i] = f[i] = pop;
      if (x < M_PI) x += dx;
      else attr *= oldageattr;
      i++;
      }
   for (i=frstage;i<=lastage;i++)
      f[i] -= f[i] * cbrate * (1.0 - cbattr);
   for (gen=0;gen<maxyrs;gen++)
      {
      i = MAXYEAR - 1;
      while (i > 0)
         {
         m[i] = m[i-1];
         f[i] = f[i-1];
         i--;
         }
      i = 0;
      x = 0;
      while (m[i] > 2.0
         && f[i] > 2.0)
         {
         if (x < M_PI)
            attr = (sinrange * sin(x)) + maxattr;
         m[i] *= attr;
         f[i] *= attr;
         if (x < M_PI) x += dx;
         else attr *= oldageattr;
         i++;
         }
      m[0] = 0.0;
      f[0] = 0.0;
      for (i=frstage;i<=lastage;i++)
         {
         m[0] += f[i] * cbrate * 0.5;
         f[0] += f[i] * cbrate * 0.5;
         f[i] -= f[i] * cbrate * (1.0 - cbattr);
         } /* for each age of birth */
      shwpop(gen,m,f);
      } /* maxyrs years */
   } /* genmdl */

void validate()
   {
   double tmpmin,tmpmax;
   double brthrate,tenmths;
   if (maxyrs == 0)
      {
      fprintf(stderr,"Missing maximum years "
         "(max_years)\n");
      errsw = 1;
   if (maxage == 0)
      {
      fprintf(stderr,"Missing maximum age "
         "(max_age)\n");
      errsw = 1;
      }
      }
   if (frstage == 0)
      {
      fprintf(stderr,"Missing first age of "
         "childbirth (first_cb_age)\n");
      errsw = 1;
      }
   if (lastage == 0)
      {
      fprintf(stderr,"Missing last age of "
         "childbirth (last_cb_age)\n");
      errsw = 1;
      }
   if (frstage > lastage)
      {
      fprintf(stderr,"First childbirth age %d "
         "should be less than "
         "last childbirth age %d\n",
         frstage, lastage);
      errsw = 1;
      }
   if (lastage > maxage)
      {
      fprintf(stderr,"Last childbirth age %d "
         "should be less than or equal to "
         "maximum age %d\n",
         lastage, maxage);
      errsw = 1;
      }
   if (frstage >= maxage)
      {
      fprintf(stderr,"First childbirth age %d "
         "should be less than "
         "maximum age %d\n",
         frstage, maxage);
      errsw = 1;
      }
   if (children == 0.0)
      {
      fprintf(stderr,"Missing number of "
         "children per female (children)\n");
      errsw = 1;
      }
   cbspan = (double) lastage - frstage + 1.0;
   tenmths = 10.0 / 12.0;
   brthrate = cbspan / children;
   if (brthrate < tenmths)
      {
      fprintf(stderr,"Birth rate of %f "
         "months should be at least 10.0\n",
         brthrate*12);
      errsw = 1;
      }
   if (startpop == 0.0)
      {
      fprintf(stderr,"Missing starting "
         "population (starting_pop)\n");
      errsw = 1;
      }
   if (minattr == 0.0)
      {
      fprintf(stderr,"Missing minimum "
         "attrition (min_attrition)\n");
      errsw = 1;
      }
   if (maxattr == 0.0)
      {
      fprintf(stderr,"Missing maximum "
         "attrition (max_attrition)\n");
      errsw = 1;
      }
   tmpmin = 1.0 - minattr;
   tmpmax = 1.0 - maxattr;
   if (tmpmin > tmpmax)
      {
      fprintf(stderr,"Minimum attrition %f "
         "should be less than "
         "maximum attrition %f\n",
         tmpmin, tmpmax);
      errsw = 1;
      }
   if (cbattr == 0.0)
      {
      fprintf(stderr,"Missing childbirth "
         "attrition (cb_attrition)\n");
      errsw = 1;
      }
   if (oldageattr == 0.0)
      {
      fprintf(stderr,"Missing old age "
         "attrition (old_age_attrition)\n");
      errsw = 1;
      }
   if (!errsw) genmdl();
   } /* validate */

int main()
   {
   int rslt;

   dbsw = 0;
   errsw = 0;

   maxage = maxyrs = frstage
      = lastage = 0;
   startpop = maxattr = minattr
      = oldageattr = prevftot
      = prevmtot = children = 0.0;

   fprintf(stderr,"mdl Version 0.4.0\n");
   fprintf(stderr,"Copyright (C) 2001"
   	" dondalah@ripco.com (Dondalah)\n\n");
   fprintf(stderr,"Mdl (Population Model) "
   	"comes with ABSOLUTELY "
   	"NO WARRANTY;\n");
   fprintf(stderr,"for details see the GNU"
   	" General Public License.\n\n");
   fprintf(stderr,"This is free software, and you"
   	" are welcome to redistribute it\n");
   fprintf(stderr,"under certain conditions;\n");
   fprintf(stderr,"for details see the GNU"
   	" General Public License.\n\n");
   fprintf(stderr,"For help browse mdl.html.\n\n");

   rslt = yyparse();
   fprintf(stderr,"Result of compile = %d\n", rslt);
   if (rslt == 0 && errsw == 0)
      validate();
   return(0);
   } /* main */
